from __future__ import absolute_import, division, print_function, unicode_literals
from iminuit import minimize
from iminuit.tests.utils import assert_allclose, requires_dependency
import numpy as np


def func(x, *args):
    c = args[0] if args else 1
    return c + x[0] ** 2 + (x[1] - 1) ** 2 + (x[2] - 2) ** 2


def grad(x, *args):
    return 2 * (x - (0, 1, 2))


@requires_dependency("scipy")
def test_simple():
    result = minimize(func, (1, 1, 1))
    assert_allclose(result.x, (0, 1, 2), atol=1e-8)
    assert_allclose(result.fun, 1)
    assert result.nfev > 0
    assert result.njev == 0


@requires_dependency("scipy")
def test_gradient():
    result = minimize(func, (1, 1, 1), jac=grad)
    assert_allclose(result.x, (0, 1, 2), atol=1e-8)
    assert_allclose(result.fun, 1)
    assert result.nfev > 0
    assert result.njev > 0


@requires_dependency("scipy")
def test_args():
    result = minimize(func, np.ones(3), args=(5,))
    assert_allclose(result.x, (0, 1, 2), atol=1e-8)
    assert_allclose(result.fun, 5)
    assert result.nfev > 0
    assert result.njev == 0


@requires_dependency("scipy")
def test_callback():
    trace = []
    result = minimize(func, np.ones(3), callback=lambda x: trace.append(x.copy()))
    assert_allclose(result.x, (0, 1, 2), atol=1e-8)
    assert_allclose(result.fun, 1)
    assert result.nfev == len(trace)
    assert_allclose(trace[0], np.ones(3), atol=1e-2)
    assert_allclose(trace[-1], result.x, atol=1e-2)



from __future__ import absolute_import, division, unicode_literals
from math import log10, floor


def format_numbers(*args):
    scales = tuple(int(round(log10(abs(x)))) if x != 0 else 1 for x in args)
    nmax = max(scales)
    nsig = min(x for x in scales if nmax - x < 9) - 1
    all_pos = True
    for arg in args:
        if arg < 0:
            all_pos = False
            break
    if nmax <= 3 and nsig >= -3:
        if nsig >= 0:
            return tuple(
                ("%i" if x < 0 or all_pos else " %i") % round(x, -nsig) for x in args
            )
        else:
            return tuple(
                (("%%.%if" if x < 0 or all_pos else " %%.%if") % min(-nsig, 3)) % x
                for x in args
            )
    result = [
        ("%%.%if" if x < 0 or all_pos else " %%.%if")
        % min(nmax - nsig, 3)
        % (x / 10 ** nmax)
        for x in args
    ]
    if nmax != 0:
        return tuple(x + "E%i" % nmax for x in result)
    return tuple(result)


def format_row(widths, *args):
    return (
        "".join(
            ("|{0:^%i}" % (w - 1) if w > 0 else "| {0:%i}" % (-w - 2)).format(a)
            for (w, a) in zip(widths, args)
        )
        + "|"
    )


def fmin(fmin):
    goaledm = 1e-4 * fmin.tolerance * fmin.up
    # despite what the doc said the code is actually 1e-4
    # http://wwwasdoc.web.cern.ch/wwwasdoc/hbook_html3/node125.html

    ws = (-32, 33)
    i1 = format_row(
        ws, "FCN = %.4G" % fmin.fval, "Ncalls=%i (%i total)" % (fmin.nfcn, fmin.ncalls)
    )
    i2 = format_row(
        ws, "EDM = %.3G (Goal: %G)" % (fmin.edm, goaledm), "up = %.1f" % fmin.up
    )
    ws = (16, 16, 12, 21)
    h1 = format_row(ws, "Valid Min.", "Valid Param.", "Above EDM", "Reached call limit")
    v1 = format_row(
        ws,
        repr(fmin.is_valid),
        repr(fmin.has_valid_parameters),
        repr(fmin.is_above_max_edm),
        repr(fmin.has_reached_call_limit),
    )
    ws = (16, 16, 12, 12, 9)
    h2 = format_row(ws, "Hesse failed", "Has cov.", "Accurate", "Pos. def.", "Forced")
    v2 = format_row(
        ws,
        repr(fmin.hesse_failed),
        repr(fmin.has_covariance),
        repr(fmin.has_accurate_covar),
        repr(fmin.has_posdef_covar),
        repr(fmin.has_made_posdef_covar),
    )

    l = len(h1) * "-"
    return "\n".join((l, i1, i2, l, h1, l, v1, l, h2, l, v2, l))


def params(mps):
    vnames = [mp.name for mp in mps]
    name_width = max([4] + [len(x) for x in vnames])
    num_width = max(2, len("%i" % (len(vnames) - 1)))

    ws = (-num_width - 2, -name_width - 3, 12, 12, 13, 13, 10, 10, 8)
    h = format_row(
        ws,
        "",
        "Name",
        "Value",
        "Hesse Err",
        "Minos Err-",
        "Minos Err+",
        "Limit-",
        "Limit+",
        "Fixed",
    )
    l = "-" * len(h)
    lines = [l, h, l]
    mes = mps.merrors
    for i, mp in enumerate(mps):
        if mes and mp.name in mes:
            me = mes[mp.name]
            val, err, mel, meu = format_numbers(mp.value, mp.error, me.lower, me.upper)
        else:
            val, err = format_numbers(mp.value, mp.error)
            mel = ""
            meu = ""
        lines.append(
            format_row(
                ws,
                str(i),
                mp.name,
                val,
                err,
                mel,
                meu,
                "%g" % mp.lower_limit if mp.lower_limit is not None else "",
                "%g" % mp.upper_limit if mp.upper_limit is not None else "",
                "yes" if mp.is_fixed else "CONST" if mp.is_const else "",
            )
        )
    lines.append(l)
    return "\n".join(lines)


def merror(me):
    mel, meu = format_numbers(me.lower, me.upper)
    stat = "Valid" if me.is_valid else "Invalid"
    summary = "| {0:^15s} | {1:^27s} |".format(me.name, stat)
    error = "| {0:^15s} | {1:^12s} | {2:^12s} |".format("Error", mel, meu)
    valid = "| {0:^15s} | {1:^12s} | {2:^12s} |".format(
        "Valid", str(me.lower_valid), str(me.upper_valid)
    )
    at_limit = "| {0:^15s} | {1:^12s} | {2:^12s} |".format(
        "At Limit", str(me.at_lower_limit), str(me.at_upper_limit)
    )
    max_fcn = "| {0:^15s} | {1:^12s} | {2:^12s} |".format(
        "Max FCN", str(me.at_lower_max_fcn), str(me.at_upper_max_fcn)
    )
    new_min = "| {0:^15s} | {1:^12s} | {2:^12s} |".format(
        "New Min", str(me.lower_new_min), str(me.upper_new_min)
    )
    hline = "-" * len(error)
    return "\n".join(
        (hline, summary, hline, error, valid, at_limit, max_fcn, new_min, hline)
    )


def matrix(m):
    n = len(m)

    is_correlation = True
    for i in range(n):
        if m[i][i] != 1.0:
            is_correlation = False
            break

    args = []
    for mi in m:
        for mj in mi:
            args.append(mj)
    nums = format_numbers(*args)

    def row_fmt(args):
        s = "| " + args[0] + " |"
        for x in args[1:]:
            s += " " + x
        s += " |"
        return s

    first_row_width = max(len(v) for v in m.names)
    row_width = max(first_row_width, max(len(v) for v in nums))
    v_names = [("{:>%is}" % first_row_width).format(x) for x in m.names]
    h_names = [("{:>%is}" % row_width).format(x) for x in m.names]
    val_fmt = ("{:>%is}" % row_width).format

    header = row_fmt([" " * first_row_width] + h_names)
    hline = "-" * len(header)
    lines = [hline, header, hline]

    for i, vn in enumerate(v_names):
        lines.append(row_fmt([vn] + [val_fmt(nums[n * i + j]) for j in range(n)]))
    lines.append(hline)
    return "\n".join(lines)



from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
from iminuit import Minuit
from iminuit.tests.utils import assert_allclose, requires_dependency

try:
    import matplotlib as mpl

    mpl.use("Agg")
except ImportError:
    pass


def f1(x, y):
    return (1 - x) ** 2 + 100 * (y - 1) ** 2


def f2(par):
    return f1(par[0], par[1])


@pytest.fixture(params=("normal", "numpy"))
def m(request):
    if request.param == "normal":
        m = Minuit(f1, x=0, y=0, pedantic=False, print_level=1)
    else:
        m = Minuit.from_array_func(
            f2, (0, 0), name=("x", "y"), pedantic=False, print_level=1
        )
    m.migrad()
    return m


@requires_dependency("matplotlib")
def test_profile(m):
    m.minos("x")
    m.draw_profile("x")


@requires_dependency("matplotlib")
def test_mnprofile(m):
    m.minos("x")
    m.draw_mnprofile("x")


@requires_dependency("matplotlib")
def test_mncontour(m):
    m.minos()
    m.draw_mncontour("x", "y")


@requires_dependency("matplotlib")
def test_drawcontour(m):
    m.minos()
    m.draw_contour("x", "y")


@requires_dependency("matplotlib")
def test_drawcontour_show_sigma(m):
    m.minos()
    m.draw_contour("x", "y", show_sigma=True)



from __future__ import absolute_import, division, print_function, unicode_literals
import warnings

__all__ = ["IMinuitWarning", "InitialParamWarning", "HesseFailedWarning"]


class IMinuitWarning(RuntimeWarning):
    """iminuit warning.
    """

    pass


class InitialParamWarning(IMinuitWarning):
    """Initial parameter warning.
    """

    pass


class HesseFailedWarning(IMinuitWarning):
    """HESSE failed warning.
    """

    pass


warnings.simplefilter("always", InitialParamWarning, append=True)



from __future__ import absolute_import, division, print_function, unicode_literals
from iminuit.color import Gradient

__all__ = ["LatexTable"]


class LatexTable:
    """Latex table output.
    """

    float_format = "%10.5g"
    int_format = "%d"
    latex_kwd = [
        "alpha",
        "beta",
        "gamma",
        "delta",
        "epsilon",
        "zeta",
        "eta",
        "theta",
        "iota",
        "kappa",
        "lambda",
        "mu",
        "nu",
        "xi",
        "omicron",
        "pi",
        "rho",
        "sigma",
        "tau",
        "upsilon",
        "phi",
        "chi",
        "psi",
        "omega",
        "Alpha",
        "Beta",
        "Gamma",
        "Delta",
        "Epsilon",
        "Zeta",
        "Eta",
        "Theta",
        "Iota",
        "Kappa",
        "Lambda",
        "Mu",
        "Nu",
        "Xi",
        "Omicron",
        "Pi",
        "Rho",
        "Sigma",
        "Tau",
        "Upsilon",
        "Phi",
        "Chi",
        "Psi",
        "Omega",
    ]

    def __init__(
        self,
        data,
        headers=None,
        smart_latex=True,
        escape_under_score=True,
        alignment=None,
        rotate_header=False,
        latex_map=None,
    ):
        # Make sure #data columns matches #header columns, using any non-zero
        # length if data or headers are missing
        if len(data) > 0:
            num_col = len(data[0])
            if headers:
                assert num_col == len(headers)
        else:
            if headers is not None:
                num_col = len(headers)
            else:
                # LaTeX requires at least one column
                num_col = 1

        self.headers = headers
        self.data = data
        self.num_col = num_col
        self.smart_latex = smart_latex
        self.escape_under_score = escape_under_score
        self.alignment = self._auto_align() if alignment is None else alignment
        self.rotate_header = rotate_header
        self.latex_map = {} if latex_map is None else latex_map
        self.cell_color = {}  # map of tuple (i,j)=>(r, g, b) #i,j include header

    def _auto_align(self):
        return "|" + "c|" * self.num_col

    def _format(self, s):
        if s in self.latex_map:
            return self.latex_map[s]
        elif isinstance(s, float):
            return self.float_format % s
        elif isinstance(s, int):
            return self.int_format % s
        elif self.smart_latex:
            return self._convert_smart_latex(s)
        elif self.escape_under_score:
            return s.replace("_", r"\_")
        else:
            return s

    def _convert_smart_latex(self, s):
        """convert greek symbol to latex one
        transform
        a to $a$ if a is greek letter else just a
        a_xxx to $a_{xxx}$ and
        a_xxx_yyy_zzz to a xxx $yyy_{zzz}$
        """
        # FIXME: implement this

        parts = s.split("_")
        if len(parts) == 1:  # a to $a$ if a is greek letter else just a
            if parts[0] in self.latex_kwd:
                return r"$\%s$" % str(parts[0])
            else:
                return str(parts[0])
        elif len(parts) == 2:  # a_xxx to $a_{xxx}$ and
            first = "\\%s" % parts[0] if parts[0] in self.latex_kwd else parts[0]
            second = "\\%s" % parts[1] if parts[1] in self.latex_kwd else parts[1]
            return r"$%s_{%s}$" % (first, second)
        else:  # a_xxx_yyy_zzz to a xxx $yyy_{zzz}$
            textpart = map(self._convert_smart_latex, parts[:-2])
            textpart = " ".join(textpart)
            latexpart = self._convert_smart_latex("_".join(parts[-2:]))
            return textpart + " " + latexpart

    def set_cell_color(self, i, j, c):
        """colorize i,j cell with rgb color tuple c

        Note that i,j index includes header.
        i=0 is header if header is present. If header is not present then
        i=0 refer to first data row.
        """
        self.cell_color[(i, j)] = c

    def _prepare(self):  # return list of list
        ret = []
        if self.headers:
            tmp = list(map(self._format, self.headers))
            if self.rotate_header:
                tmp = list(map(lambda x: "\\rotatebox{90}{%s}" % x, tmp))

            ret.append(tmp)
        for x in self.data:
            ret.append(list(map(self._format, x)))
        return ret

    def __str__(self):
        hline = "\\hline\n"
        ret = ""
        if len(self.cell_color) != 0:
            ret += "%\\usepackage[table]{xcolor} % include this for color\n"
            ret += "%\\usepackage{rotating} % include this for rotate header\n"
            ret += "%\\documentclass[xcolor=table]{beamer} % for beamer\n"
        ret += "\\begin{tabular}{%s}\n" % self.alignment
        ret += hline
        tdata = self._prepare()
        # decorate it

        for (i, j), c in self.cell_color.items():
            xcolor = "[RGB]{%d,%d,%d}" % (c[0], c[1], c[2])
            tdata[i][j] = "\\cellcolor" + xcolor + " " + tdata[i][j]

        for line in tdata:
            ret += " & ".join(line) + "\\\\\n"
            ret += hline
        ret += "\\end{tabular}\n"

        return ret.strip()


class LatexFactory:
    @classmethod
    def build_matrix(cls, vnames, matrix, latex_map=None):
        """build latex correlation matrix"""
        # ret_link  = '<a onclick="$(\'#%s\').toggle()" href="#">Show Latex</a>'%uid
        headers = [""] + list(vnames)
        data = []
        color = {}
        grad = Gradient(
            (-1.0, 120.0, 120.0, 250.0),
            (0.0, 250.0, 250.0, 250.0),
            (1.0, 250.0, 100.0, 100.0),
        )
        for i, v1 in enumerate(vnames):
            tmp = [v1]
            for j, v2 in enumerate(vnames):
                m = matrix[i][j]
                tmp.append(m)
                color[(i + 1, j + 1)] = grad(m)
                # +1 for header on the side and top
            data.append(tmp)

        table = LatexTable(
            headers=headers, data=data, rotate_header=True, latex_map=latex_map
        )
        table.float_format = "%.2g"
        for (i, j), c in color.items():
            table.set_cell_color(i, j, c)
        return table

    @classmethod
    def build_param_table(
        cls, mps, merr=None, float_format="%5.3e", smart_latex=True, latex_map=None
    ):
        """build latex parameter table"""
        headers = [
            "",
            "Name",
            "Value",
            "Hesse Error",
            "Minos Error-",
            "Minos Error+",
            "Limit-",
            "Limit+",
            "Fixed?",
        ]

        data = []
        for i, mp in enumerate(mps):
            minos_p, minos_m = (
                ("", "")
                if merr is None or mp.name not in merr
                else ("%g" % merr[mp.name].upper, "%g" % merr[mp.name].lower)
            )
            limit_p = "" if mp.upper_limit is None else "%g" % mp.upper_limit
            limit_m = "" if mp.lower_limit is None else "%s" % mp.lower_limit
            fixed = "Yes" if mp.is_fixed else "No"
            tmp = [
                i,
                mp.name,
                "%g" % mp.value,
                "%g" % mp.error,
                minos_m,
                minos_p,
                limit_m,
                limit_p,
                fixed,
            ]
            data.append(tmp)
        alignment = "|c|r|r|r|r|r|r|r|c|"
        ret = LatexTable(
            data,
            headers=headers,
            alignment=alignment,
            smart_latex=smart_latex,
            latex_map=latex_map,
        )
        ret.float_format = float_format
        return ret
